{-# LANGUAGE OverloadedStrings #-}

module Windows12.Parser (programP) where

import Control.Monad.Combinators.Expr
import Text.Megaparsec
import Windows12.Ast
import Windows12.Lexer

opTable :: [[Operator Parser Expr]]
opTable =
  [ [ InfixL $ Member <$ symbol ".",
      InfixL $ (\l r -> Member (UnOp Deref l) r) <$ symbol "->"
    ],
    [ unary (UnOp Neg) "-",
      unary (UnOp Not) "!",
      unary (UnOp BitNot) "~",
      unary (UnOp Deref) "*",
      unary (UnOp AddrOf) "&"
    ],
    [ Postfix $ flip Index <$> (symbol "[" *> exprP <* symbol "]")
    ],
    [ Postfix $ flip Call <$> parens (sepBy exprP (symbol ","))
    ],
    [ infixL' Mul "*",
      infixL' Div "/",
      infixL' Mod "%"
    ],
    [ infixL' Add "+",
      infixL' Sub "-"
    ],
    [ infixL' ShiftL "<<",
      infixL' ShiftR ">>"
    ],
    [ infixL Le "<=",
      infixL Lt "<",
      infixL Ge ">=",
      infixL Gt ">"
    ],
    [ infixL Eq "==",
      infixL Ne "!="
    ],
    [ infixL' BitAnd "&"
    ],
    [ infixL' BitXor "^"
    ],
    [ infixL' BitOr "|"
    ],
    [ infixL And "&&"
    ],
    [ infixL Or "||"
    ]
  ]
  where
    unary op sym = Prefix $ foldr1 (.) <$> some (op <$ symbol sym)
    infixL op sym = InfixL $ BinOp op <$ symbol sym
    infixL' op sym = InfixL $ BinOp op <$ operator sym
    infixR op sym = InfixR $ BinOp op <$ symbol sym
    operator sym = lexeme $ try $ (symbol sym <* notFollowedBy opChar)
    opChar = oneOf ("+-*/%<>&|^=!~" :: [Char])

termP :: Parser Expr
termP =
  parens exprP
    <|> IntLit <$> intLiteral
    <|> UIntLit <$> uintLiteral
    <|> try (FloatLit <$> floatLiteral)
    <|> StrLit <$> stringLiteral
    <|> BoolLit <$> (reserved "true" *> pure True <|> reserved "false" *> pure False)
    <|> CharLit <$> charLiteral
    <|> try (Sizeof <$> (reserved "sizeof" *> typeP))
    <|> try (Cast <$> (parens typeP) <*> termP)
    <|> try (StructInit <$> identifier <*> braces (sepEndBy1 ((,) <$> identifier <* symbol ":" <*> exprP) (symbol ",")))
    <|> Id <$> identifier

exprP :: Parser Expr
exprP = makeExprParser termP opTable

structP :: Parser TLStruct
structP = do
  reserved "struct"
  name <- identifier
  fields <- braces (sepEndBy (Bind <$> identifier <* symbol ":" <*> typeP) (symbol ","))
  return $ Struct name fields

typeP :: Parser Type
typeP = do
  t <-
    ArrayType <$> (brackets typeP)
      <|> IntType <$ reserved "int"
      <|> UIntType <$ reserved "uint"
      <|> FloatType <$ reserved "float"
      <|> StrType <$ reserved "str"
      <|> BoolType <$ reserved "bool"
      <|> CharType <$ reserved "char"
      <|> StructType <$> identifier
  foldr (const PtrType) t <$> many (symbol "*")

assignP :: Parser Stmt
assignP = do
  lhs <- exprP
  op <-
    AddAssign <$ symbol "+="
      <|> SubAssign <$ symbol "-="
      <|> MulAssign <$ symbol "*="
      <|> DivAssign <$ symbol "/="
      <|> ModAssign <$ symbol "%="
      <|> BitAndAssign <$ symbol "&="
      <|> BitOrAssign <$ symbol "|="
      <|> BitXorAssign <$ symbol "^="
      <|> ShiftLAssign <$ symbol "<<="
      <|> ShiftRAssign <$ symbol ">>="
      <|> BaseAssign <$ symbol "="
  Assign op lhs <$> exprP <* symbol ";"

stmtP :: Parser Stmt
stmtP =
  Return <$> (reserved "return" *> exprP <* symbol ";")
    <|> (If <$> (reserved "if" *> exprP) <*> braces (many stmtP) <*> optional (reserved "else" *> braces (many stmtP)))
    <|> (While <$> (reserved "while" *> parens exprP) <*> braces (many stmtP))
    <|> (Var <$> (reserved "var" *> identifier) <*> optional (symbol ":" *> typeP) <*> optional (symbol "=" *> exprP) <* symbol ";")
    <|> try assignP
    <|> Expr <$> exprP <* symbol ";"
    <|> Block <$> braces (many stmtP)

funcP :: Parser TLFunc
funcP = do
  reserved "fn"
  name <- identifier
  args <- parens (sepBy (Bind <$> identifier <* symbol ":" <*> typeP) (symbol ","))
  retType <- (symbol "->" *> typeP) <|> pure VoidType
  body <- braces (many stmtP)
  return $ Func name args retType body

enumP :: Parser TLEnum
enumP = do
  reserved "enum"
  name <- identifier
  fields <- braces (sepEndBy1 identifier (symbol ","))
  return $ Enum name fields

memberFuncP :: Parser TLFunc
memberFuncP = do
  reserved "fn"
  reserved "on"
  self <- typeP
  name <- identifier
  args <- parens (sepBy (Bind <$> identifier <* symbol ":" <*> typeP) (symbol ","))
  retType <- (symbol "->" *> typeP) <|> pure VoidType
  body <- braces (many stmtP)
  return $ Func name (Bind "self" self : args) retType body

organize :: [TL] -> Program
organize tls = Program structs enums funcs
  where
    structs = [s | TLStruct s <- tls]
    enums = [e | TLEnum e <- tls]
    funcs = [f | TLFunc f <- tls]

programP :: Parser Program
programP = between sc eof $ do
  tls <- many (TLStruct <$> structP <|> TLEnum <$> enumP <|> try (TLFunc <$> memberFuncP) <|> TLFunc <$> funcP)
  return $ organize tls